"""
Data processing functionality.

.. sidebar:: Processing vs. analysis steps

    The key difference between processing and analysis steps: While a
    processing step *modifies* the data of the dataset it operates on,
    an analysis step returns a result based on data of a dataset, but leaves
    the original dataset unchanged.


Key to reproducible science is automatic documentation of each processing
step applied to the data of a dataset. Such a processing step each is
self-contained, meaning it contains every necessary information to perform
the processing task on a given dataset.

Processing steps, in contrast to analysis steps (see :mod:`trepr.analysis`
for details), not only operate on data of a :class:`trepr.dataset.Dataset`,
but change its data. The information necessary to reproduce each processing
step gets added to the :attr:`trepr.dataset.Dataset.history` attribute of a
dataset.

Due to the inheritance from the :mod:`aspecd.processing` module, all processing
steps provided are fully self-documenting, i.e. they add all necessary
information to reproduce each processing step to the
:attr:`trepr.dataset.Dataset.history` attribute of the dataset.


Concrete processing steps
=========================

This module provides a series of processing steps that can be divided into
those specific for TREPR data and those generally applicable to
spectroscopic data and directly inherited from the ASpecD framwork.

What follows is a list as a first overview. For details, see the detailed
documentation of each of the classes, readily accessible by the link.


Processing steps specific for TREPR data
----------------------------------------

A number of processing steps are rather specific for TREPR data namely
correcting DC offsets, background, and microwave frequency:

* :class:`PretriggerOffsetCompensation`

  Correct for DC offsets of TREPR data

* :class:`BackgroundCorrection`

  Subtract background, mainly laser-induced field-independent background

* :class:`FrequencyCorrection`

  Correct for same microwave frequency, necessary to compare measurements


General processing steps inherited from the ASpecD framework
------------------------------------------------------------

Besides the processing steps specific for TREPR data, a number of further
processing steps that are generally applicable to spectroscopic data have
been inherited from the underlying ASpecD framework:

* :class:`Normalisation`

  Normalise data.

  There are different kinds of normalising data: maximum, minimum,
  amplitude, area

* :class:`ScalarAlgebra`

  Perform scalar algebraic operation on one dataset.

  Operations available: add, subtract, multiply, divide (by given scalar)

* :class:`ScalarAxisAlgebra`

  Perform scalar algebraic operation on axis values of a dataset.

  Operations available: add, subtract, multiply, divide, power (by given scalar)

* :class:`DatasetAlgebra`

  Perform scalar algebraic operation on two datasets.

  Operations available: add, subtract

* :class:`Projection`

  Project data, *i.e.* reduce dimensions along one axis.

* :class:`SliceExtraction`

  Extract slice along one ore more dimensions from dataset.

* :class:`BaselineCorrection`

  Correct baseline of dataset.

* :class:`Averaging`

  Average data over given range along given axis.

* :class:`Filtering`

  Filter data


Further processing steps implemented in the ASpecD framework can be used as
well, by importing the respective modules. In case of recipe-driven data
analysis, simply prefix the kind with ``aspecd``:

.. code-block:: yaml

    - kind: aspecd.processing
      type: <ClassNameOfProcessingStep>


Implementing own processing steps is rather straight-forward. For details,
see the documentation of the :mod:`aspecd.processing` module.

Module documentation
====================

"""
import numpy as np
import scipy.signal

import aspecd.processing
import aspecd.exceptions

import trepr.exceptions


class PretriggerOffsetCompensation(aspecd.processing.SingleProcessingStep):
    # noinspection PyUnresolvedReferences
    """
    Correct for DC offsets of TREPR data.

    Usually the first processing step after recording TREPR data is to
    compensate for DC offsets due to experimental instabilities. This is
    done by setting the average of the pretrigger part of the time trace to
    zero (pretrigger offset compensation). At the same time, this will
    remove any background signals of stable paramangetic species, as they
    would appear as DC offset as well.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        zeropoint_index : :class:`int`
            Index of the time axis corresponding to *t* = 0

            Will be automatically detected during processing.


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the pretrigger offset compensation with
    default values:

    .. code-block:: yaml

       - kind: processing
         type: PretriggerOffsetCompensation

    This will correct your data accordingly and should always be the first
    step when processing and analysing TREPR data.

    """

    def __init__(self):
        super().__init__()
        # public properties:
        # Note: self.parameters inherited
        self.description = 'Pretrigger offset compensation'
        self.undoable = True
        self.parameters['zeropoint_index'] = 0

    def _perform_task(self):
        """Perform the processing step and return the processed data."""
        self._get_zeropoint_index()
        self._execute_compensation(self.parameters['zeropoint_index'])

    def _get_zeropoint_index(self):
        """Get the index of the last time value before the trigger."""
        zeropoint_index = \
            np.argmin(abs(self.dataset.data.axes[1].values))
        self.parameters['zeropoint_index'] = int(zeropoint_index)

    def _execute_compensation(self, range_end):
        """Execute the pretrigger offset compensation."""
        for field_point, time_trace in enumerate(self.dataset.data.data):
            pretrig_avg = self._get_pretrigger_average(time_trace, range_end)
            self.dataset.data.data[field_point] = time_trace - pretrig_avg

    @staticmethod
    def _get_pretrigger_average(time_trace, range_end=1):
        """Calculate the average of the time trace before triggering."""
        array = time_trace[0:range_end]
        return np.average(array)


class BackgroundCorrection(aspecd.processing.SingleProcessingStep):
    # noinspection PyUnresolvedReferences
    """
    Subtract background, mainly laser-induced field-independent background.

    When the laser hits the EPR cavity, this usually introduces a
    field-independent absorptive background signal that needs to be
    subtracted from the data.

    Depending on the spectrometer control and measurement software used,
    this background signal can get automatically subtracted already during
    the measurement. More often, it needs to be done afterwards,
    and therefore, it is crucial to record the TREPR data with sufficient
    baseline at both ends of the magnetic field range to allow for reliable
    background correction.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        num_profiles : :class:`list`
            Number of time profiles (transients) to use from lower and upper
            end of the magnetic field axis.

            If two values are provided, a linear regression will be
            performed between lower and upper end and the background
            subtracted accordingly. If only a scalar (or a list with one
            element) is provided, the background traces from the lower
            magnetic field position are used.

            Default: [5, 5]


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the background correction with default
    values:

    .. code-block:: yaml

       - kind: processing
         type: BackgroundCorrection

    This will correct your data accordingly.

    If you would like to control more carefully the transients (time profiles)
    used to obtain the background signal, you can set the respective
    parameters. Suppose you would want to use only the first 10 transients
    from the lower end of the magnetic field:

    .. code-block:: yaml

       - kind: processing
         type: BackgroundCorrection
         properties:
           parameters:
             num_profiles: 10

    Similarly, if you would want to use only the *last* 10 transients from the
    lower end of the magnetic field:

    .. code-block:: yaml

       - kind: processing
         type: BackgroundCorrection
         properties:
           parameters:
             num_profiles: -10

    And finally, if you would like to use the first 5 and the last 10
    transients, you would write:

    .. code-block:: yaml

       - kind: processing
         type: BackgroundCorrection
         properties:
           parameters:
             num_profiles: [5, 10]

    """

    def __init__(self):
        super().__init__()
        # public properties:
        self.description = 'Background correction of 2D spectrum'
        self.undoable = True
        self.parameters['num_profiles'] = [5, 5]

    @staticmethod
    def applicable(dataset):
        """
        Check whether processing step is applicable to the given dataset.

        Background correction is only applicable to 2D datasets.

        Parameters
        ----------
        dataset : :class:`aspecd.dataset.Dataset`
            dataset to check

        Returns
        -------
        applicable : :class:`bool`
            `True` if successful, `False` otherwise.

        """
        return len(dataset.data.axes) == 3

    def _sanitise_parameters(self):
        self.parameters['num_profiles'] = list(self.parameters['num_profiles'])
        if len(self.parameters['num_profiles']) == 1:
            self.parameters['num_profiles'] = self.parameters['num_profiles'][0]

    def _perform_task(self):
        self._check_data_size()
        self._subtract_background()

    def _check_data_size(self):
        if isinstance(self.parameters['num_profiles'], list):
            num_profiles = sum(self.parameters['num_profiles'])
        else:
            num_profiles = self.parameters['num_profiles']
        if len(self.dataset.data.axes[0].values) <= 2 * num_profiles:
            raise aspecd.exceptions.NotApplicableToDatasetError(
                message='The given dataset ist too small to perform '
                        'appropriate background correction.')

    def _subtract_background(self):
        if isinstance(self.parameters['num_profiles'], list) and \
                len(self.parameters['num_profiles']) == 2:
            self._bg_corr_with_slope()
        else:
            self._bg_corr_one_side()

    def _bg_corr_with_slope(self):
        low = self.parameters['num_profiles'][0]
        high = abs(self.parameters['num_profiles'][1])
        lower_mean = np.mean(self.dataset.data.data[:low])
        higher_mean = np.mean(self.dataset.data.data[-high:])
        slope = (higher_mean - lower_mean) / self.dataset.data.data.shape[0]
        for idx, transient in enumerate(self.dataset.data.data):
            transient -= lower_mean + slope * idx

    def _bg_corr_one_side(self):
        assert isinstance(self.parameters['num_profiles'], int)

        if self.parameters['num_profiles'] < 0:
            self._subtract_from_end()
        else:
            self._subtract_from_begin()

    def _subtract_from_end(self):
        bg = np.mean(self.dataset.data.data[self.parameters['num_profiles']:])
        self.dataset.data.data -= bg

    def _subtract_from_begin(self):
        bg = np.mean(self.dataset.data.data[:self.parameters['num_profiles']])
        self.dataset.data.data -= bg


class FrequencyCorrection(aspecd.processing.SingleProcessingStep):
    # noinspection PyUnresolvedReferences
    """Convert data to a given microwave frequency.

    To compare EPR spectra, it is necessary to first correct them for the
    same microwave frequency, *i.e.* to adjust the magnetic field axis
    accordingly. Note that each individual measurement will have its own
    microwave frequency. Particularly for TREPR data with their usually
    quite large steps of the magnetic field axis, one could first check
    whether the difference in microwave frequency is reasonably large
    compared to the magnetic field steps, and only in this case correct for
    the same frequency.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        frequency : :class:`float`
            Microwave frequency to correct for in GHz.

        Default: 9.5


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the frequency correction with default
    values:

    .. code-block:: yaml

       - kind: processing
         type: FrequencyCorrection

    This will correct your data accordingly.

    If you would like to set the target microwave frequency explicitly,
    this can be done as well:

    .. code-block:: yaml

       - kind: processing
         type: BackgroundCorrection
         properties:
           parameters:
             frequency: 9.8

    In this case, the data would be corrected for a microwave frequency of
    9.8 GHz.

    .. codeauthor:: Mirjam Schröder

    """

    def __init__(self):
        super().__init__()
        self.parameters["frequency"] = 9.5
        self.description = "Correct magnetic field axis for given frequency"

    def _perform_task(self):
        """Perform the actual transformation / correction.

        For the conversion the x axis data is first converted to an axis in
        units of using the given frequency, then converted back using target
        frequency.
        """
        nu_target = self.parameters['frequency']
        for axis in self.dataset.data.axes:
            # TODO: Question: Better check for quantity rather than unit? (
            #   Difficult if not filled)
            # if axis.quantity == 'magnetic field'
            if axis.unit in ('mT', 'G'):
                axis.values = self._correct_field_for_frequency(nu_target,
                                                                axis.values)
                self._write_new_frequency()

    def _correct_field_for_frequency(self, nu_target=None,
                                     b_initial=None):
        """
        Calculate new field axis for given frequency.

        Parameters
        ----------
        nu_target : :class:`float`
            Frequency the magnetic field should be computed for

        b_initial : :class:`numpy.ndarray`
            Original field axis

        Returns
        -------
        b_target : :class:`numpy.ndarray`
            Computed field axis

        """
        nu_initial = self.dataset.metadata.bridge.mw_frequency.value
        b_target = (nu_target / nu_initial) * b_initial
        return b_target

    def _write_new_frequency(self):
        self.dataset.metadata.bridge.mw_frequency.value = \
            self.parameters['frequency']


class Normalisation(aspecd.processing.Normalisation):
    """Normalise data.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.Normalisation`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. Of course, all parameters settable
    for the superclasses can be set as well. The examples focus each on a
    single aspect.

    In the simplest case, just invoke the normalisation with default values:

    .. code-block:: yaml

       - kind: processing
         type: Normalisation

    This will normalise your data to their maximum.

    Sometimes, normalising to maximum is not what you need, hence you can
    control in more detail the criterion using the appropriate parameter:

    .. code-block:: yaml

       - kind: processing
         type: Normalisation
         properties:
           parameters:
             kind: amplitude

    In this case, you would normalise to the amplitude, meaning setting the
    difference between minimum and maximum to one. For other kinds, see above.

    If you want to normalise not over the entire range of the dataset,
    but only over a dedicated range, simply provide the necessary parameters:

    .. code-block:: yaml

       - kind: processing
         type: Normalisation
         properties:
           parameters:
             range: [50, 150]

    In this case, we assume a 1D dataset and use indices, requiring the data
    to span at least over 150 points. Of course, it is often more convenient
    to provide axis units. Here you go:

    .. code-block:: yaml

       - kind: processing
         type: Normalisation
         properties:
           parameters:
             range: [340, 350]
             range_unit: axis

    And in case of ND datasets with N>1, make sure to provide as many ranges
    as dimensions of your dataset, in case of a 2D dataset:

    .. code-block:: yaml

       - kind: processing
         type: Normalisation
         properties:
           parameters:
             range:
               - [50, 150]
               - [30, 40]

    Here as well, the range can be given in indices or axis units,
    but defaults to indices if no unit is explicitly given.

    """


class ScalarAlgebra(aspecd.processing.ScalarAlgebra):
    """Perform scalar algebraic operation on one dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.ScalarAlgebra`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In case you would like to add a fixed value of 42 to your dataset:

    .. code-block:: yaml

       - kind: processing
         type: ScalarAlgebra
         properties:
           parameters:
             kind: add
             value: 42

    Similarly, you could use "minus", "times", "by", "add", "subtract",
    "multiply", or "divide" as kind - resulting in the given algebraic
    operation.

    """


class ScalarAxisAlgebra(aspecd.processing.ScalarAxisAlgebra):
    """Perform scalar algebraic operation on the axis of a dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.ScalarAxisAlgebra`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In case you would like to add a fixed value of 42 to the first axis
    (index 0) your dataset:

    .. code-block:: yaml

       - kind: processing
         type: ScalarAxisAlgebra
         properties:
           parameters:
             kind: plus
             axis: 0
             value: 42

    Similarly, you could use "minus", "times", "by", "add", "subtract",
    "multiply", "divide", and "power" as kind - resulting in the given
    algebraic operation.

    """


class DatasetAlgebra(aspecd.processing.DatasetAlgebra):
    """Perform scalar algebraic operation on two datasets.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.DatasetAlgebra`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In case you would like to add the data of the dataset referred to by its
    label ``label_to_other_dataset`` to your dataset:

    .. code-block:: yaml

       - kind: processing
         type: DatasetAlgebra
         properties:
           parameters:
             kind: plus
             dataset: label_to_other_dataset

    Similarly, you could use "minus", "add", "subtract" as kind - resulting
    in the given algebraic operation.

    As mentioned already, the data of both datasets need to have identical
    shape, and comparison is only meaningful if the axes are compatible as
    well. Hence, you will usually want to perform a CommonRangeExtraction
    processing step before doing algebra with two datasets:

    .. code-block:: yaml

       - kind: multiprocessing
         type: CommonRangeExtraction
         results:
           - label_to_dataset
           - label_to_other_dataset

       - kind: processing
         type: DatasetAlgebra
         properties:
           parameters:
             kind: plus
             dataset: label_to_other_dataset
         apply_to:
           - label_to_dataset

    """


class Projection(aspecd.processing.Projection):
    """Project data, *i.e.* reduce dimensions along one axis.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.Projection`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the projection with default values:

    .. code-block:: yaml

       - kind: processing
         type: Projection

    This will project the data along the first axis (index 0), yielding a 1D
    dataset.

    If you would like to project along the second axis (index 1), simply set
    the appropriate parameter:

    .. code-block:: yaml

       - kind: processing
         type: Projection
         properties:
           parameters:
             axis: 1

    This will project the data along the second axis (index 1), yielding a 1D
    dataset.

    """


class SliceExtraction(aspecd.processing.SliceExtraction):
    """Extract slice along one ore more dimensions from dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.SliceExtraction`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the slice extraction with an index only:

    .. code-block:: yaml

       - kind: processing
         type: SliceExtraction
         properties:
           parameters:
             position: 5

    This will extract the sixth slice (index five) along the first axis (index
    zero).

    If you would like to extract a slice along the second axis (with index
    one), simply provide both parameters, index and axis:

    .. code-block:: yaml

       - kind: processing
         type: SliceExtraction
         properties:
           parameters:
             position: 5
             axis: 1

    This will extract the sixth slice along the second axis.

    And as it is sometimes more convenient to give ranges in axis values
    rather than indices, even this is possible. Suppose the axis you would
    like to extract a slice from runs from 340 to 350 and you would like to
    extract the slice corresponding to 343:

    .. code-block:: yaml

       - kind: processing
         type: SliceExtraction
         properties:
           parameters:
             position: 343
             unit: axis

    In case of you providing the range in axis units rather than indices,
    the value closest to the actual axis value will be chosen automatically.

    For ND datasets with N>2, you can either extract a 1D or ND slice,
    with N always at least one dimension less than the original data. To
    extract a 2D slice from a 3D dataset, simply proceed as above, providing
    one value each for position and axis. If, however, you want to extract a
    1D slice from a 3D dataset, you need to provide two values each for
    position and axis:

    .. code-block:: yaml

       - kind: processing
         type: SliceExtraction
         properties:
           parameters:
             position: [21, 42]
             axis: [0, 2]

    This particular case would be equivalent to ``data[21, :, 42]`` assuming
    ``data`` to contain the numeric data, besides, of course, that the
    processing step takes care of removing the axes as well.

    """


class BaselineCorrection(aspecd.processing.BaselineCorrection):
    """Subtract baseline from dataset.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.BaselineCorrection`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    In the simplest case, just invoke the baseline correction with default
    values:

    .. code-block:: yaml

       - kind: processing
         type: BaselineCorrection

    In this case, a zeroth-order polynomial baseline will be subtracted from
    your dataset using ten percent to the left and right, and in case of a
    2D dataset, the baseline correction will be performed along the first
    axis (index zero) for all indices of the second axis (index 1).

    Of course, often you want to control a little bit more how the baseline
    will be corrected. This can be done by explicitly setting some parameters.

    Suppose you want to perform a baseline correction with a polynomial of
    first order:

    .. code-block:: yaml

       - kind: processing
         type: BaselineCorrection
         properties:
           parameters:
             order: 1

    If you want to change the (percental) area used for fitting the
    baseline, and even specify different ranges left and right:

    .. code-block:: yaml

       - kind: processing
         type: BaselineCorrection
         properties:
           parameters:
             fit_area: [5, 20]

    Here, five percent from the left and 20 percent from the right are used.

    Finally, suppose you have a 2D dataset and want to average along the
    second axis (index one):

    .. code-block:: yaml

       - kind: processing
         type: BaselineCorrection
         properties:
           parameters:
             axis: 1

    Of course, you can combine the different options.

    """


class Filtering(aspecd.processing.Filtering):
    """Filter data.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.processing.Filtering`
    class for details.

    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Generally, filtering requires to provide both, a type of filter and a
    window length. Therefore, for uniform and Gaussian filters, this would be:

    .. code-block:: yaml

       - kind: processing
         type: Filtering
         properties:
           parameters:
             type: uniform
             window_length: 10

    Of course, at least uniform filtering (also known as boxcar or moving
    average) is strongly discouraged due to the artifacts introduced.
    Probably the best bet for applying a filter to smooth your data is the
    Savitzky-Golay filter:

    .. code-block:: yaml

       - kind: processing
         type: Filtering
         properties:
           parameters:
             type: savitzky-golay
             window_length: 10
             order: 3

    Note that for this filter, you need to provide the polynomial order as
    well. To get best results, you will need to experiment with the
    parameters a bit.

    """


class Averaging(aspecd.processing.SingleProcessingStep):
    # noinspection PyUnresolvedReferences
    """Averaging of two-dimensional data along a given axis.

    When measuring TREPR data, the resulting spectrum is always
    two-dimensional. To analyse the data it's often necessary to extract a
    one-dimensional spectrum. The one-dimensional spectrum can either be a cut
    along the field or the time axis. To get a representative spectrum, the
    average over several points along the respective axis is calculated.

    All parameters, implicit and explicit, necessary to perform the averaging
    processing step, will be stored in the attribute
    :attr:`trepr.processing.Averaging.parameters`.

    An example for using the averaging processing step may look like this::

        avg = Averaging(dimension=0, avg_range=[4.e-7, 6.e-7], unit='axis')
        dataset.process(avg)

    Parameters
    ----------
    dimension : {0,1}, optional
        Dimension along which the averaging is done. 0 is along the field axis
        and 1 is along the time axis. Default is 0.

    avg_range : list
        Range in which the averaging will take place.

    unit : {'axis', 'index'}, optional
        Unit in which the average range is given. Either 'axis' for axis values
        or 'index' for indices. Default is axis.

    Attributes
    ----------
    parameters : dict
        All Parameters that must be known to reproduce the averaging.

    description : str
        Describes the aim of the class.

    undoable : bool
        Information weather the processing step is undoable or not.

    Raises
    ------
    trepr.exceptions.DimensionError
        Raised if dimension is not in [0, 1].

    trepr.exceptions.UnitError
        Raised if unit is not in ['axis', 'index'].

    trepr.exceptions.RangeError
        Raised if range is not within axis.


    .. deprecated:: 0.1
        Use :class:`aspecd.processing.Averaging` instead.

    """

    def __init__(self, dimension=0, avg_range=None, unit='axis'):
        super().__init__()
        # public properties
        self.description = 'Averaging'
        self.undoable = True
        self.parameters['dimension'] = dimension
        self.parameters['range'] = avg_range
        self.parameters['unit'] = unit
        # protected properties:
        self._dim = 1 if self.parameters['dimension'] == 0 else 0

    def _perform_task(self):
        """Perform the processing step."""
        avg_range = self._get_avg_range()
        self._execute_averaging(avg_range)

    def _get_avg_range(self):
        """Calculate the range within the averaging will be performed."""
        if self.parameters['unit'] == 'axis':
            start_index = \
                self._get_index(self.dataset.data.axes[self._dim].values,
                                self.parameters['range'][0])
            end_index = \
                self._get_index(self.dataset.data.axes[self._dim].values,
                                self.parameters['range'][-1])
            avg_range = [start_index, end_index]
        else:
            avg_range = self.parameters['range']
        return avg_range

    def _execute_averaging(self, avg_range):
        """Apply the averaging on the given experimental dataset."""
        if self._dim == 0:
            # keep time and intensity axes
            axes = [self.dataset.data.axes[1], self.dataset.data.axes[2]]
            self.dataset.data.data = np.average(self.dataset.data.data[:,
                                                avg_range[0]:avg_range[1]],
                                                axis=0)
        else:
            # keep B and intensity axes
            axes = [self.dataset.data.axes[0], self.dataset.data.axes[2]]
            self.dataset.data.data = np.average(self.dataset.data.data[:,
                                                avg_range[0]:avg_range[1]],
                                                axis=1)
        self.dataset.data.axes = axes

    @staticmethod
    def _get_index(vector, value):
        return np.abs(vector - value).argmin()

    def _sanitise_parameters(self):
        if self.parameters['dimension'] not in [0, 1]:
            raise trepr.exceptions.DimensionError(
                'Wrong dimension. Choose 0 or 1.')
        if self.parameters['unit'] not in ['index', 'axis']:
            raise trepr.exceptions.UnitError(
                'Wrong unit. Choose "axis" or "index".')
        if self.parameters['unit'] == 'index':
            if self.parameters['range'][0] not in \
                    range(len(self.dataset.data.axes[self._dim].values)):
                raise trepr.exceptions.RangeError('Lower index out of range.')
            if self.parameters['range'][1] not in \
                    range(len(self.dataset.data.axes[self._dim].values)):
                raise trepr.exceptions.RangeError('Upper index out of range.')
        if self.parameters['unit'] == 'axis':
            if not self._value_within_vector_range(
                    self.parameters['range'][0],
                    self.dataset.data.axes[self._dim].values):
                raise trepr.exceptions.RangeError('Lower value out of range.')
            if not self._value_within_vector_range(
                    self.parameters['range'][1],
                    self.dataset.data.axes[self._dim].values):
                raise trepr.exceptions.RangeError('Upper value out of range.')
        if self.parameters['range'][1] < self.parameters['range'][0]:
            raise trepr.exceptions.RangeError('Values need to be ascending.')

    @staticmethod
    def _value_within_vector_range(value, vector):
        return np.amin(vector) <= value <= np.amax(vector)


class Filter(aspecd.processing.SingleProcessingStep):
    # noinspection PyUnresolvedReferences
    """Apply a filter to smooth 1D data.

    Be careful to show filtered spectra.

    It can be chosen between boxcar, Savitzky-Golay and binomial filters.

    Savitzky-Golay:
    Takes a certain number of points and fits a polynomial through them.

    Reference for the Savitzky-Golay Filter:

     * A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of Data by
       Simplified Least Squares Procedures. Analytical Chemistry, 1964,
       36 (8), pp 1627-1639.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        type : :class:`str`
            Type of the applied filter. Valid inputs: savitzky-golay,
            binomial, boxcar and some abbreviations and variations.

            Default: savitzky-golay

        window_width : :class:`int`
            Full filter window width. Must be an odd number

            Default: 1/5 of the data length.


    .. deprecated:: 0.1
        Use :class:`Filtering` instead.

    """

    def __init__(self):
        super().__init__()
        self.description = 'Filter 1D dataset.'
        self.parameters['type'] = 'savitzky-golay'
        self.parameters['window_width'] = None
        self.parameters['order'] = 2

    @staticmethod
    def applicable(dataset):
        """
        Check whether processing step is applicable to the given dataset.

        Filtering is only applicable to 1D datasets.

        Parameters
        ----------
        dataset : :class:`aspecd.dataset.Dataset`
            dataset to check

        Returns
        -------
        applicable : :class:`bool`
            `True` if successful, `False` otherwise.

        """
        return len(dataset.data.axes) == 2

    def _sanitise_parameters(self):
        self._set_defaults()  # must be here for referring to right type.
        if int(self.parameters['window_width']) % 2 == 0:
            self.parameters['window_width'] += 1
            print('For applying the filter, the window length must be odd. '
                  'I added one.')

    def _get_type(self):
        """Allow for different inputs, unify them."""
        types = {
            'savitzky_golay': ['savitzky_golay', 'savitzky-golay',
                               'savitzky golay', 'savgol', 'savitzky'],
            'binomial': ['binom', 'binomial'],
            'boxcar': ['box', 'boxcar', 'moving-average', 'car']
        }
        for key, value in types.items():
            if self.parameters['type'] in value:
                self.parameters['type'] = key
            if self.parameters['type'] == 'car':
                print('Haha, good joke! You\'ve got a boxcar')

    def _perform_task(self):
        if self.parameters['type'] == 'boxcar':
            self._apply_boxcar()
        elif self.parameters['type'] == 'binomial':
            self._apply_binomial()
        else:
            self._apply_savitzky_golay()

    def _set_defaults(self):
        self._get_type()
        if not self.parameters['window_width']:
            self.parameters['window_width'] = int(np.ceil((1 / 10 * len(
                self.dataset.data.axes[0].values))) * 2 + 1)

    def _apply_savitzky_golay(self):
        filtered_data = \
            scipy.signal.savgol_filter(self.dataset.data.data,
                                       int(self.parameters['window_width']),
                                       self.parameters['order'])
        self.dataset.data.data = filtered_data

    def _apply_binomial(self):
        self._add_padding()
        self._perform_binomial_filtering()

    def _apply_boxcar(self):
        self._add_padding()
        self._perform_boxcar_filtering()

    def _add_padding(self):
        """Add padding to get same length of data at the end."""
        width = self.parameters['window_width']
        self.dataset.data.data = np.concatenate((
            np.ones(int(np.floor(width / 2))) * self.dataset.data.data[0],
            self.dataset.data.data,
            np.ones(int(np.floor(width / 2)) + 1) * self.dataset.data.data[-1]
        ))

    def _perform_binomial_filtering(self):
        filter_coefficients = (np.poly1d([0.5, 0.5]) ** self.parameters[
            'window_width']).coeffs
        filtered_data = np.array(np.convolve(self.dataset.data.data,
                                             filter_coefficients, mode='valid'))
        self.dataset.data.data = filtered_data

    def _perform_boxcar_filtering(self):
        filter_ = np.ones(self.parameters['window_width']) / self.parameters[
            'window_width']
        filtered_data = np.array(np.convolve(self.dataset.data.data,
                                             filter_, mode='valid'))
        self.dataset.data.data = filtered_data[:-1]



"""
General facilities for input (and output).

In order to work with TREPR data, these data need to be imported into the
trepr package. Therefore, the module provides importers for specific file
formats. In case of TREPR spectroscopy, the measurement control software is
often lab-written and specific for a local setup. One exception is the
Bruker BES3T file format written by Bruker Xepr and Xenon software that can
be used to record TREPR data in combination with a pulsed EPR spectrometer.

Another class implemented in this module is the
:class:`trepr.io.DatasetImporterFactory`, a prerequisite for recipe-driven
data analysis. This factory returns the correct dataset importer for a
specific dataset depending on the information provided (usually, a filename).


Importers for specific file formats
===================================

Currently, two file formats are supported by specific importers:

* :class:`SpeksimImporter`

  The Speksim format was developed in Freiburg in the group of Prof. G.
  Kothe and used afterwards in the group of Prof. S. Weber. The spectrometer
  control software was developed by Prof. U. Heinen.

  One speciality of this file format is that each transient is stored in an
  individual file. For each of these transients, a timestamp as well as the
  microwave frequency are recorded as well, allowing to analyse frequency
  drifts and irregularities in the data acquisition.

* :class:`TezImporter`

  The tez file format is the internal format used by the MATLAB(r) trepr
  toolbox developed by T. Biskup. It vaguely resembles the OpenDocument
  format used, *e.g.*, by OpenOffice and LibreOffice. In short, the metadata
  are contained in an XML file, while the numerical data are stored as IEEE
  754 standard binaries in separate files. The ASpecD dataset format (adf)
  is similar in some respect.


Implementing importers for additional file formats is rather
straight-forward. For details, see the documentation of the :mod:`aspecd.io`
module.


Module documentation
====================

"""
import collections
import datetime
import glob
import io
import os
import re
import shutil
from zipfile import ZipFile

import numpy as np
import xmltodict

import aspecd.annotation
import aspecd.io
import aspecd.infofile
import aspecd.metadata
import aspecd.plotting
import aspecd.utils

import trepr.dataset


class DatasetImporterFactory(aspecd.io.DatasetImporterFactory):
    """Factory for creating importer objects based on the source provided.

    Often, data are available in different formats, and deciding which
    importer is appropriate for a given format can be quite involved. To
    free other classes from having to contain the relevant code, a factory
    can be used.

    Currently, the sole information provided to decide about the appropriate
    importer is the source (a string). A concrete importer object is
    returned by the method ``get_importer()``. If no source is provided,
    an exception will be raised.

    If the source string does not match any of the importers handled by this
    module, the standard importers from the ASpecD framework are checked.
    See the documentation of the :class:`aspecd.io.DatasetImporterFactory`
    base class for details.

    """

    def _get_importer(self):
        if os.path.isdir(self.source):
            return SpeksimImporter(source=self.source)
        return TezImporter(source=self.source)


class SpeksimImporter(aspecd.io.DatasetImporter):
    """Import trepr raw data in Freiburg Speksim format including its metadata.

    Datasets in this format consist of several time traces, each of which is
    stored in a text file. In order to analyse the raw data, it is necessary
    to store the time traces all together in one dataset.

    The corresponding metadata are read from an external file in infofile
    format. For further information about the infofile format see:
    `<https://www.till-biskup.de/en/software/info/format>`_.

    Parameters
    ----------
    source : :class:`str`
        Path to the raw data.

    Attributes
    ----------
    dataset : :obj:`trepr.dataset.ExperimentalDataset`
        Entity containing data and metadata.

    Raises
    ------
    FileNotFoundError
        Raised if no infofile could be found.

    """

    def __init__(self, source=''):
        super().__init__(source=source)
        # public properties
        self.dataset = None
        # protected properties
        self._headerlines = 5
        self._data = np.array([])
        self._file_format = str()
        self._time_stamps = np.array([])
        self._mwfreq = np.array([])
        self._comment_line = str()
        self._time_unit = str()
        self._field_unit = str()
        self._intensity_unit = str()
        self._mwfreq_unit = str()
        self._time_axis = np.array([])
        self._field_axis = np.array([])
        self._infofile = aspecd.infofile.Infofile()
        self._header = list()
        self._format_no = int()
        self._time_start = float()
        self._time_stop = float()
        self._time_points = int()

    def _import(self):
        """Execute all necessary methods and write the data to a dataset."""
        self._import_raw_data()
        self._hand_data_to_dataset()

        self._create_time_axis()
        self._ensure_field_axis_in_SI_unit()
        self._hand_axes_to_dataset()

        self._load_infofile()
        self._map_infofile()

        self._create_time_stamp_data()
        self._create_mw_freq_data()

    def _import_raw_data(self):
        """Import the time traces and cut off the header lines."""
        filenames = self._get_filenames()
        for filename in filenames:
            self._process_timetrace(filename)
        self._data = \
            np.reshape(self._data, [len(filenames), self._time_points])

    def _get_filenames(self):
        filenames = sorted(glob.glob(os.path.join(self.source,
                                                  '*.[0-9][0-9][0-9]')))
        return filenames

    def _process_timetrace(self, filename):
        with open(filename) as file:
            raw_data = file.read()
        lines = raw_data.splitlines()
        self._header = lines[0:self._headerlines]
        self._parse_header()
        numeric_data = self._process_numeric_data(raw_data)
        self._data = np.append(self._data, numeric_data)

    def _process_numeric_data(self, raw_data):
        # noinspection PyTypeChecker
        numeric_data = np.loadtxt(io.StringIO(raw_data),
                                  skiprows=self._headerlines)
        numeric_data = np.reshape(numeric_data, self._time_points)
        return numeric_data

    def _parse_header(self):
        """Execute the methods which parse the header lines."""
        self._parse_header_1st_line()
        self._parse_header_2nd_line()
        self._parse_header_3rd_line()
        self._parse_header_4th_line()
        self._parse_header_5th_line()

    def _parse_header_1st_line(self):
        """Parse the 1st header line and extract the time and date.

        Example::

            Source : transient; Time : Wed Jun 7 08:44:57 2017

        """
        entries = self._header[0].split(';')
        self._file_format = entries[0].split(':')[1].strip()
        time_stamp = entries[1].split(' : ')[1]
        time_stamp = datetime.datetime.strptime(time_stamp,
                                                '%a %b %d %H:%M:%S %Y')
        # noinspection PyTypeChecker
        self._time_stamps = np.append(self._time_stamps, time_stamp)

    def _parse_header_2nd_line(self):
        """Extract the field and frequency unit from the 2nd header line.

        Example::

            B0 = 4080.000000 Gauss, mw = 9.684967 GHz

        """

        def parse_line(line):
            matches = re.search('([A-Za-z0-9]*) = ([0-9.]*) ([A-Za-z]*)', line)
            return float(matches.group(2)), matches.group(3)

        entries = self._header[1].split(',')
        field, self._field_unit = parse_line(entries[0])
        mwfreq, self._mwfreq_unit = parse_line(entries[1])
        self._field_axis = np.append(self._field_axis, field)
        self._mwfreq = np.append(self._mwfreq, mwfreq)

    def _parse_header_3rd_line(self):
        """Parse the 3rd header line.

        Example::

             NDI-T2 sa64 20/2 42/25 523nm/1mJ

        """
        self._comment_line = self._header[2]

    def _parse_header_4th_line(self):
        """Extract format number and time information from 4th header line.

        Example::

            1 5000 -1.001e-06 8.997e-06 0 0

        """
        entries = self._header[3].split()[0:4]
        self._format_no = int(entries[0])
        self._time_points = int(entries[1])
        self._time_start = float(entries[2])
        self._time_stop = float(entries[3])

    def _parse_header_5th_line(self):
        """Extract the time- and intensity-unit from the 5th header line.

        Example::

            s                        V

        """
        self._time_unit, self._intensity_unit = self._header[4].split()

    def _create_time_axis(self):
        """Create the time axis using the start, end, and time points."""
        self._time_axis = \
            np.linspace(self._time_start,
                        self._time_stop,
                        num=self._time_points)

    def _load_infofile(self):
        """Import the infofile and parse it."""
        infofile_name = self._get_infofile_name()
        if not infofile_name:
            raise FileNotFoundError('Infofile not found.')
        self._infofile.filename = infofile_name[0]
        self._infofile.parse()

    def _get_infofile_name(self):
        return glob.glob(os.path.join(self.source, '*.info'))

    def _map_infofile(self):
        """Bring the metadata to a given format."""
        infofile_version = self._infofile.infofile_info['version']
        self._map_metadata(infofile_version)
        self._assign_comment_as_annotation()

    def _assign_comment_as_annotation(self):
        comment = aspecd.annotation.Comment()
        comment.comment = self._infofile.parameters['COMMENT']
        self.dataset.annotate(comment)

    def _map_metadata(self, infofile_version):
        """Bring the metadata into a unified format."""
        mapper = aspecd.metadata.MetadataMapper()
        mapper.version = infofile_version
        mapper.metadata = self._infofile.parameters
        root_path = os.path.split(os.path.abspath(__file__))[0]
        mapper.recipe_filename = os.path.join(root_path, 'metadata_mapper.yaml')
        mapper.map()
        self.dataset.metadata.from_dict(mapper.metadata)

    def _hand_data_to_dataset(self):
        """Hand the data to the dataset structure."""
        self.dataset.data.data = self._data

    # noinspection PyPep8Naming
    def _ensure_field_axis_in_SI_unit(self):  # noqa: N802
        """Ensure that the field axis unit is in SI unit."""
        if self._field_unit == 'Gauss':
            self._field_unit = 'mT'
            self._field_axis = self._field_axis / 10

    def _hand_axes_to_dataset(self):
        """Hand the axes and intensity to the dataset structure."""
        self.dataset.data.axes[0].values = self._field_axis
        self.dataset.data.axes[0].unit = self._field_unit
        self.dataset.data.axes[0].quantity = 'magnetic field'
        self.dataset.data.axes[1].values = self._time_axis
        self.dataset.data.axes[1].unit = self._time_unit
        self.dataset.data.axes[1].quantity = 'time'
        self.dataset.data.axes[2].unit = self._intensity_unit
        self.dataset.data.axes[2].quantity = 'intensity'

    def _create_time_stamp_data(self):
        """Hand the time stamp data to the dataset structure."""
        self.dataset.time_stamp.data = self._time_stamps
        self.dataset.time_stamp.axes[0].values = self._field_axis
        self.dataset.time_stamp.axes[0].unit = self._field_unit
        self.dataset.time_stamp.axes[0].quantity = 'magnetic field'
        self.dataset.time_stamp.axes[1].quantity = 'date'

    def _create_mw_freq_data(self):
        """Hand the microwave frequency data to the dataset structure."""
        self.dataset.microwave_frequency.data = self._mwfreq
        self.dataset.microwave_frequency.axes[0].values = self._field_axis
        self.dataset.microwave_frequency.axes[0].unit = self._field_unit
        self.dataset.microwave_frequency.axes[0].quantity = 'magnetic field'
        self.dataset.microwave_frequency.axes[1].unit = self._mwfreq_unit
        self.dataset.microwave_frequency.axes[1].quantity = \
            'microwave frequency'


class TezImporter(aspecd.io.DatasetImporter):
    """Importer for MATLAB(r) trepr toolbox format.

    The MATLAB(r) trepr toolbox format is basically a ZIP archive consisting
    of a list of standard IEEE 754 binary files containing the data and an
    XML file containing the accompanying metadata in structured form,
    enriched with information necessary to directly convert them back into
    MATLAB structures (corresponding to a Python :class:`dict`).

    Parameters
    ----------
    source : :class:`str`
        Path to the raw data.

    Attributes
    ----------
    dataset : :obj:`trepr.dataset.ExperimentalDataset`
        Entity containing data and metadata.

    """

    def __init__(self, source=''):
        # Dirty fix: Cut file extension
        if source.endswith(".tez"):
            source = source[:-4]
        super().__init__(source=source)
        # public properties
        self.tez_mapper_filename = 'tez_mapper.yaml'
        self.xml_dict = None
        self.dataset = None
        self.metadata_filename = ''
        self.load_infofile = True
        # private properties
        self._metadata = None
        self._infofile = aspecd.infofile.Infofile()
        self._root_dir = ''
        self._filename = ''
        self._tmpdir = ''
        self._raw_data_name = ''
        self._raw_data_shape_filename = ''

    def _import(self):
        self._unpack_zip()
        self._get_dir_and_filenames()
        self._import_xml_data_to_dict()
        self._get_data_from_binary()
        self._parse_axes()

        if self.load_infofile and self._infofile_exists():
            self._load_infofile()
            self._map_infofile()

        self._get_metadata_from_xml()
        self._get_mw_frequencies()

        self._remove_tmp_directory()

    def _unpack_zip(self):
        self._root_dir, self._filename = os.path.split(self.source)
        self._tmpdir = os.path.join(self._root_dir, 'tmp')
        with ZipFile(self.source + '.tez', 'r') as zip_obj:
            zip_obj.extractall(self._tmpdir)

    def _get_dir_and_filenames(self):
        hidden_filename = os.listdir(os.path.join(self._root_dir, 'tmp'))[0]
        self.metadata_filename = os.path.join(self._root_dir, 'tmp',
                                              hidden_filename, 'struct.xml')
        self._raw_data_name = \
            os.path.join(self._root_dir, 'tmp', hidden_filename,
                         'binaryData', 'data')
        self._raw_data_shape_filename = os.path.join(self._raw_data_name +
                                                     '.dim')

    def _import_xml_data_to_dict(self):
        with open(self.metadata_filename, 'r') as file:
            xml_data = file.read()
        self.xml_dict = xmltodict.parse(xml_data)

    def _get_data_from_binary(self):
        with open(self._raw_data_shape_filename, 'r') as f:
            shape = list([int(x) for x in f.read().split()])
        shape.reverse()  # Shape is given in reverse order in .dim file
        raw_data = np.fromfile(self._raw_data_name, dtype='<f8')
        raw_data = np.reshape(raw_data, shape).transpose()
        self.dataset.data.data = raw_data

    def _parse_axes(self):
        if len(self.xml_dict['struct']['axes']['data']['measure']) > 3:
            raise NotImplementedError('No method to import more than 3 axes. '
                                      'This task is left to you.')
        for axis in self.xml_dict['struct']['axes']['data']['measure']:
            self._get_magnetic_field_axis(axis)
            self._get_time_axis(axis)

    def _get_magnetic_field_axis(self, axis):
        if '#text' in axis.keys() and axis['#text'] == 'magnetic field':
            id_ = int(axis['@id']) - 1
            self.dataset.data.axes[0].quantity = 'magnetic field'
            self.dataset.data.axes[0].values = \
                self._get_values_from_xml_dict(id_=id_)
            assert int(self.xml_dict['struct']['axes']['data']['values'][
                id_]['@id']) == (id_ + 1), 'Axis-IDs do not match!'
            self.dataset.data.axes[0].unit = self.xml_dict['struct']['axes'][
                'data']['unit'][id_]['#text']

    def _get_time_axis(self, axis):
        if '#text' in axis.keys() and axis['#text'] == 'time':
            id_ = int(axis['@id']) - 1
            self.dataset.data.axes[1].quantity = 'time'
            self.dataset.data.axes[1].values = \
                self._get_values_from_xml_dict(id_=id_)
            assert int(self.xml_dict['struct']['axes']['data']['values'][
                id_]['@id']) == (id_ + 1)
            self.dataset.data.axes[1].unit = self.xml_dict['struct']['axes'][
                'data']['unit'][id_]['#text']

    def _infofile_exists(self):
        if self._get_infofile_name() and os.path.exists(
                self._get_infofile_name()[0]):
            return True
        print('No infofile found for dataset %s, import continued without '
              'infofile.' % os.path.split(self.source)[1])
        return False

    def _get_infofile_name(self):
        return glob.glob(''.join([self.source.strip(), '.info']))

    def _load_infofile(self):
        """Import infofile and parse it."""
        infofile_name = self._get_infofile_name()
        self._infofile.filename = infofile_name[0]
        self._infofile.parse()

    def _map_infofile(self):
        """Bring the metadata to a given format."""
        infofile_version = self._infofile.infofile_info['version']
        self._map_metadata(infofile_version)
        self._assign_comment_as_annotation()

    def _map_metadata(self, infofile_version):
        """Bring the metadata into a unified format."""
        mapper = aspecd.metadata.MetadataMapper()
        mapper.version = infofile_version
        mapper.metadata = self._infofile.parameters
        root_path = os.path.split(os.path.abspath(__file__))[0]
        mapper.recipe_filename = os.path.join(root_path, 'metadata_mapper.yaml')
        mapper.map()
        self._metadata = \
            aspecd.utils.convert_keys_to_variable_names(mapper.metadata)

    def _assign_comment_as_annotation(self):
        comment = aspecd.annotation.Comment()
        comment.comment = self._infofile.parameters['COMMENT']
        self.dataset.annotate(comment)

    def _get_values_from_xml_dict(self, id_=None):
        values = np.asarray([float(i) for i in
                             self.xml_dict['struct']['axes']['data'][
                                 'values'][id_]['#text'].split(' ') if i])
        return values

    def _get_metadata_from_xml(self):
        mapping = aspecd.utils.Yaml()
        rootpath = os.path.split(os.path.abspath(__file__))[0]
        mapping.read_from(os.path.join(rootpath, self.tez_mapper_filename))
        metadata_dict = collections.OrderedDict()
        for key, subdict in mapping.dict.items():
            metadata_dict[key] = collections.OrderedDict()
            for key2, value in subdict.items():
                metadata_dict[key][key2] = \
                    self._cascade(self.xml_dict['struct'], value)

        self._metadata = self._fuse_with_existing_metadata(metadata_dict)
        self.dataset.metadata.from_dict(self._metadata)
        # Cause Copycat in UdS measurement program:
        self.dataset.metadata.bridge.attenuation.unit = 'dB'

    def _fuse_with_existing_metadata(self, metadata_dict):
        metadata_dict = \
            aspecd.utils.remove_empty_values_from_dict(metadata_dict)
        infofile_metadata = \
            aspecd.utils.copy_values_between_dicts(target=self._metadata,
                                                   source=metadata_dict)
        return infofile_metadata

    def _cascade(self, dict_, value):
        keys = value.split('.')
        return_value = dict_
        for key in keys:
            return_value = return_value[key]
        if self._get_physical_quantity(return_value):
            return_value = self._get_physical_quantity(return_value)
        elif self._get_value(return_value):
            return_value = self._get_value(return_value)
        else:
            return_value = ''
        return return_value

    @staticmethod
    def _get_value(dict_):
        return_value = None
        if '#text' in dict_.keys():
            return_value = dict_['#text']
        return return_value

    @staticmethod
    def _get_physical_quantity(dict_):
        return_value = None
        if 'value' and 'unit' in dict_.keys():
            if '#text' in dict_['value'].keys():
                return_value = {
                    'value': float(dict_['value']['#text']),
                    'unit': dict_['unit']['#text']
                }
        return return_value

    def _get_mw_frequencies(self):
        """Get the dataset with real frequencies of each magnetic field point.

        This is special for the trepr dataset but useful to track frequency
        drifts. In th UdS-measurement program, the frequency is automaticly
        written in the tez structure.
        """
        if self._xml_contains_mw_frequencies():
            self.dataset.microwave_frequency.data = \
                np.asarray([float(i) for i in self.xml_dict['struct'][
                    'parameters']['bridge']['MWfrequency']['values'][
                    '#text'].split(' ') if i])
            self.dataset.microwave_frequency.axes[0] = \
                self.dataset.data.axes[0]
            self.dataset.microwave_frequency.axes[1].unit = \
                self.dataset.metadata.bridge.mw_frequency.unit
            self.dataset.microwave_frequency.axes[1].quantity = \
                'microwave frequency'

    def _xml_contains_mw_frequencies(self):
        answer = False
        if '#text' in self.xml_dict['struct']['parameters']['bridge'][
                'MWfrequency']['values']:
            answer = True
        return answer

    def _remove_tmp_directory(self):
        if os.path.exists(self._tmpdir):
            shutil.rmtree(self._tmpdir)



"""
trepr package.

Package for reproducible and traceable analysis of trepr data.

Available modules
-----------------
:mod:`trepr.dataset`
    Organise dataset, consisting of data and metadata, and the metadata.
:mod:`trepr.io`
    Import trepr raw data in different formats including its metadata.
:mod:`trepr.processing`
    Process data in datasets.
:mod:`trepr.analysis`
    Analyse data in datasets.
:mod:`trepr.plotting`
    Plot data in datasets.
:mod:`trepr.report`
    Create reports with given datasets.

Utilities
---------
:mod:`trepr.exceptions`
    Exceptions for the trepr package


"""



"""
Exceptions for the trepr package.

For preventing cyclic imports and for a better overview, all exception
classes of the trepr package are collected in this module. It is save for
every other module to import this module, as this module does *not* depend
on any other modules.

"""


class Error(Exception):
    """Base class for exceptions in this module."""


class DimensionError(Error):
    """Exception raised when the dimension of the data does not fit.

    Attributes
    ----------
    message : `str`
        explanation of the error

    """

    def __init__(self, message=''):
        super().__init__()
        self.message = message


class RangeError(Error):
    """Exception raised when the given range is out of the dataset's range.

    Attributes
    ----------
    message : `str`
        explanation of the error

    """

    def __init__(self, message=''):
        super().__init__()
        self.message = message


class UnitError(Error):
    """Exception raised when the unit does not fit.

    Attributes
    ----------
    message : `str`
        explanation of the error

    """

    def __init__(self, message=''):
        super().__init__()
        self.message = message


"""
Data analysis functionality.

.. sidebar:: Processing vs. analysis steps

    The key difference between processing and analysis steps: While a
    processing step *modifies* the data of the dataset it operates on,
    an analysis step returns a result based on data of a dataset, but leaves
    the original dataset unchanged.


Key to reproducible science is automatic documentation of each analysis
step applied to the data of a dataset. Such an analysis step each is
self-contained, meaning it contains every necessary information to perform
the analysis task on a given dataset.

Analysis steps, in contrast to processing steps (see
:mod:`trepr.processing` for details), operate on data of a
:class:`trepr.dataset.Dataset`, but don't change its data. Rather,
some result is obtained that is stored separately, together with the
parameters of the analysis step, in the
:attr:`trepr.dataset.Dataset.analyses` attribute of the dataset.

In order to quantify the quality of a measured spectrum or to interpret it, it
is often necessary to perform some analysis steps.

Due to inheritance from the :mod:`aspecd.analysis` module all analysis steps
provided are fully self-documenting, i.e. they add all necessary information
to reproduce each analysis step to the :attr:`aspecd.dataset.Dataset.history`
attribute of the dataset.

"""

import datetime
import matplotlib.dates as mdates

import numpy as np
import scipy.constants

import aspecd.analysis
import aspecd.metadata
import aspecd.utils


class MwFreqAnalysis(aspecd.analysis.SingleAnalysisStep):
    """
    Calculate the frequency drift and compare it with the step size.

    In order to estimate the quality of a spectrum, it can be helpful to know
    the extent the frequency drifted during the measurement.

    An example for using the microwave frequency analysis step may look like
    this::

        dataset_ = trepr.dataset.ExperimentalDataset()
        analysis_step = MwFreqAnalysis()
        dataset_.analyse(analysis_step)

    Attributes
    ----------
    description : str
        Describes the aim of the class.

    """

    def __init__(self):
        super().__init__()
        # public properties
        self.description = 'Microwave frequency drift analysis.'
        # protected properties
        self._delta_mw_freq = float()
        self._delta_B0 = float()
        self._step_size_in_mT = float()
        self._ratio_frequency_drift_to_step_size = float()

    @staticmethod
    def applicable(dataset):
        """Check whether processing step is applicable to the given dataset."""
        return dataset.microwave_frequency.data.any()

    def _perform_task(self):
        """Perform all methods to do analysis."""
        self._calculate_mw_freq_amplitude()
        self._convert_delta_mw_freq_to_delta_B0()
        self._calculate_step_size()
        self._compare_delta_B0_with_step_size()
        self._write_result()

    def _calculate_mw_freq_amplitude(self):
        """Calculate the amplitude of the microwave frequency."""
        self._delta_mw_freq = max(self.dataset.microwave_frequency.data) - \
            min(self.dataset.microwave_frequency.data)

    def _convert_delta_mw_freq_to_delta_B0(self):  # noqa: N802
        """Calculate delta B0 by using the resonance condition."""
        electron_g_factor = scipy.constants.value('electron g factor')
        bohr_magneton = scipy.constants.value('Bohr magneton')
        planck_constant = scipy.constants.value('Planck constant')
        self._delta_B0 = self._delta_mw_freq * 1e9 * planck_constant \
            / (-1 * electron_g_factor * bohr_magneton)

    def _calculate_step_size(self):
        """Calculate the step size of the given dataset."""
        self._step_size_in_mT = \
            self.dataset.microwave_frequency.axes[0].values[1] \
            - self.dataset.microwave_frequency.axes[0].values[0]

    def _compare_delta_B0_with_step_size(self):  # noqa: N802
        """Calculate the ratio between delta B0 and the step size."""
        self._ratio_frequency_drift_to_step_size = \
            self._delta_B0 / self._step_size_in_mT

    def _write_result(self):
        """Write the results in the results dictionary."""
        self.result = {
            'frequency drift': aspecd.metadata.PhysicalQuantity(
                value=self._delta_B0, unit='T'),
            'ratio frequency drift/step size':
                self._ratio_frequency_drift_to_step_size}


class TimeStampAnalysis(aspecd.analysis.SingleAnalysisStep):
    """
    Calculate the time spent for recording each time trace.

    Can be helpful for debugging the spectrometer.

    An example for using the time stamp analysis step may look like
    this::

        dataset_ = trepr.dataset.ExperimentalDataset()
        analysis_step = TimeStampAnalysis()
        dataset_.analyse(analysis_step)

    Attributes
    ----------
    description : str
        Describes the aim of the class.

    """

    def __init__(self):
        super().__init__()
        # public properties
        self.description = 'Time stamp analysis.'
        # protected properties
        self._time_field_matrix = None
        self._time_stamp_datetimes = list()

    def _perform_task(self):
        """Perform all methods to do analysis."""
        self._create_time_field_matrix()
        self._calculate_time_stamp_delta()
        self._write_result()

    def _create_time_field_matrix(self):
        """Create matrix with time stamps and corresponding field points."""
        time_stamp_floats = np.zeros(0)
        for time_stamp in self.dataset.time_stamp.data:
            time_stamp_floats = \
                np.append(time_stamp_floats, time_stamp.timestamp())
        self._time_field_matrix = np.zeros((len(time_stamp_floats), 2))
        self._time_field_matrix[:, 0] = time_stamp_floats
        self._time_field_matrix[:, 1] = self.dataset.time_stamp.axes[0].values
        self._time_field_matrix = \
            self._time_field_matrix[self._time_field_matrix[:, 0].argsort()]
        for time_stamp in self._time_field_matrix[:, 0]:
            self._time_stamp_datetimes.append(
                datetime.datetime.fromtimestamp(time_stamp))

    def _calculate_time_stamp_delta(self):
        """Calculate the time between the time stamps."""
        zero = datetime.datetime(2018, 1, 1)
        for i in range(len(self._time_stamp_datetimes) - 1):
            self._time_stamp_datetimes[i] = \
                self._time_stamp_datetimes[i + 1] - \
                self._time_stamp_datetimes[i] + zero
        del self._time_stamp_datetimes[-1]
        zero = mdates.date2num(zero)
        for i in range(len(self._time_stamp_datetimes)):
            self._time_stamp_datetimes[i] = \
                mdates.date2num(self._time_stamp_datetimes[i]) - zero
        self._time_field_matrix = \
            np.delete(self._time_field_matrix, -1, axis=0)

    def _write_result(self):
        """Assign result.

        The result is assigned to :attr:`aspecd.analysis.AnalysisStep.result`.
        """
        self.result = {'time spent per time trace': self._time_stamp_datetimes}


"""
General facilities to generate a report.

To do scientific research in terms of reproducibility and traceability it's
highly necessary to report all the steps done on a given dataset and never
separate the dataset from its metadata.

This module provides functionality to create a report containing a 1D and 2D
plot of the dataset, its metadata and all processing steps including
parameters.
"""

import collections
import time

import jinja2

import aspecd.report


class LaTeXReporter(aspecd.report.LaTeXReporter):
    """
    Generate a report based on a LaTeX template provided.

    An example for using the :class:`trepr.report.LaTeXReporter` class may
    look like this::

        template_ = 'path/to/your/template.tex'
        report_output = 'path/to/your/report_output.tex'
        dataset_ = trepr.dataset.ExperimentalDataset()
        report = LaTeXReporter(template_, report_output)
        report.dataset = dataset_
        report.create()
        report.compile()

    Parameters
    ----------
    template : str
        Path to template file used to generate report.

    filename : str
        Path of the resulting template file.

    Attributes
    ----------
    dataset : :obj:`trepr.dataset.Dataset`
        Dataset structure containing raw dat as well as metadata.

    """

    def __init__(self, template='', filename=''):
        # public properties
        super().__init__(template=template, filename=filename)
        self.dataset = trepr.dataset.Dataset()
        # protected properties
        self._latex_jinja_env = jinja2.Environment()
        self._metadata = dict()
        self._date = None
        self._processing_steps = collections.OrderedDict()
        self._avg_parameter = dict()
        self._figure_name = dict()

    def create(self):
        """Perform all methods to generate a report."""
        self._prepare_metadata()
        self._get_processing_steps()
        self._get_figure_names()
        self._get_current_date()
        self._create_context()
        super().create()

    def _prepare_metadata(self):
        """Prepare the metadata the way it can be rendered automatically."""
        self._metadata = self.dataset.metadata.to_dict()
        self._metadata = self._change_keys_in_dict_recursively(self._metadata)
        self._metadata['Parameter'] = collections.OrderedDict()
        self._collect_parameters()

    def _collect_parameters(self):
        """Collect all the metadata keys."""
        for key in self._metadata.keys():
            if key not in ['Sample', 'Measurement', 'Parameter']:
                self._metadata['Parameter'][key] = \
                    self._metadata[key]

    def _get_processing_steps(self):
        """Get processing steps from history."""
        for history_record in self.dataset.history:
            self._processing_steps[history_record.processing.description] = \
                history_record.processing.parameters
        self._processing_steps = \
            self._change_keys_in_dict_recursively(self._processing_steps)

    def _change_keys_in_dict_recursively(self, dict_=None):
        """Replace all underscores in the keys with a space.

        Note: This is done because LaTeX interprets the underscore not as
        underscore but as command for subscription.
        """
        tmp_dict = collections.OrderedDict()
        for key, value in dict_.items():
            if isinstance(value, dict):
                dict_[key] = \
                    self._change_keys_in_dict_recursively(value)
            tmp_dict[key.replace('_', ' ').capitalize()] = dict_[key]
        return tmp_dict

    def _get_figure_names(self):
        """Get the names of the figures used for the report."""
        for i in range(len(self.dataset.representations)):
            if self.dataset.representations[i].plot.description \
                    == '2D plot as scaled image.':
                self._figure_name['Figure2D'] = \
                    self.dataset.representations[i].plot.filename
            elif self.dataset.representations[i].plot.description \
                    == '1D line plot.':
                self._figure_name['Figure1D'] = \
                    self.dataset.representations[i].plot.filename
            else:
                pass

    def _get_current_date(self):
        """Get the current date."""
        self._date = time.strftime('%d.%m.%Y')

    def _create_context(self):
        """Create a dictionary containing all data to write the report."""
        self.context['PROCESSINGSTEPS'] = self._processing_steps
        self.context['METADATA'] = self._metadata
        self.context['DATE'] = self._date
        self.context['PROCESSINGPARAMETERS'] = self._avg_parameter
        self.context['FIGURENAMES'] = self._figure_name


if __name__ == '__main__':
    import trepr.dataset
    import trepr.io
    import trepr.processing

    template_ = 'report.tex'
    filename_ = 'test.tex'
    dataset = trepr.dataset.ExperimentalDataset()
    imp = trepr.io.SpeksimImporter(
        source='/home/popp/nas/Python/Daten/messung01')
    dataset.import_from(imp)
    poc = trepr.processing.PretriggerOffsetCompensation()
    dataset.process(poc)
    pro = trepr.processing.Averaging(
        avg_range=[4.e-7, 6.e-7], dimension=0, unit='axis')
    dataset.process(pro)
    report = LaTeXReporter(template_, filename_)
    report.dataset = dataset
    report.create()
    report.compile()



from distutils.core import setup
import setuptools
import os

with open(os.path.join(os.path.dirname(__file__), 'VERSION')) as version_file:
    version = version_file.read().strip()

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    readme = f.read()

setup(
    name='trepr',
    version=version,
    description='TREPR processing and analysis routines.',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Jara Popp, Till Biskup, Mirjam Schröder',
    author_email='till@till-biskup.de',
    url='https://www.trepr.de/',
    project_urls={
        "Documentation": "https://docs.trepr.de/",
        "Source": "https://github.com/tillbiskup/trepr",
    },
    packages=setuptools.find_packages(exclude=('tests', 'docs')),
    include_package_data=True,
    keywords=[
        'spectroscopy', 
        'time-resolved electron paramagnetic resonance',
        'trepr',
        'data processing and analysis'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
    ],
    install_requires=[
        'aspecd>=0.2.1',
        'jinja2',
        'matplotlib',
        'numpy',
        'scipy',
        'xmltodict',
        'colour',
    ],
    extras_require={
        'dev': ['prospector'],
        'docs': ['sphinx', 'sphinx_rtd_theme'],
    },
    python_requires='>=3.5',
)



"""Datasets: units containing data and metadata.

The dataset is one key concept of the ASpecD framework and hence the trepr
package derived from it, consisting of the data as well as the corresponding
metadata. Storing metadata in a structured way is a prerequisite for a
semantic understanding within the routines. Furthermore, a history of every
processing, analysis and annotation step is recorded as well, aiming at a
maximum of reproducibility. This is part of how the ASpecD framework and
therefore the trepr package tries to support good scientific practice.

Therefore, each processing and analysis step of data should always be
performed using the respective methods of a dataset, at least as long as it
can be performed on a single dataset.


Datasets
========

Generally, there are two types of datasets: Those containing experimental
data and those containing calculated data. Therefore, two corresponding
subclasses exist:

  * :class:`trepr.dataset.ExperimentalDataset`
  * :class:`trepr.dataset.CalculatedDataset`


Metadata
========

Furthermore, in this module, the individual metadata classes are defined
which contain the individual information about the experiment:

  * :class:`trepr.dataset.ExperimentalDatasetMetadata`
  * :class:`trepr.dataset.CalculatedDatasetMetadata`

What may sound like a minor detail is one key aspect of the trepr package:
The metadata and their structure provide a unified interface for all
functionality operating on datasets. Furthermore, the metadata contained
particularly in the :class:`trepr.dataset.ExperimentalDatasetMetadata` class
are the result of more than fifteen years of practical experience. Reproducible
research is only possible if all information necessary is always recorded,
and this starts with all the metadata accompanying a measurement. Defining
what kind of metadata is important and needs to be recorded, together with
metadata formats easily writable by the experimenters *during* recording the
data requires a thorough understanding of both, the method and the setup(s)
used. For an overview of the structures of the dataset classes and their
corresponding metadata, see the :doc:`dataset structure </dataset-structure>`
section.


Dataset factory
===============

Particularly in case of recipe-driven data analysis (c.f. :mod:`aspecd.tasks`),
there is a need to automatically retrieve datasets using nothing more than a
source string that can be, e.g., a path or LOI. This is where the
DatasetFactory comes in. This is a factory in the sense of the factory
pattern described by the "Gang of Four" in their seminal work, "Design
Patterns" (Gamma et al., 1995):

  * :class:`trepr.dataset.DatasetFactory`


Module documentation
====================

"""
import os

import aspecd.dataset
import aspecd.metadata
import aspecd.utils
import trepr.io


class Dataset(aspecd.dataset.Dataset):
    """Base class for all kinds of datasets.

    Generally, there are two types of datasets: Those containing
    experimental data and those containing calculated data. Therefore,
    two corresponding subclasses exist:

      * :class:`trepr.dataset.ExperimentalDataset`
      * :class:`trepr.dataset.CalculatedDataset`

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.dataset.Dataset` class for
    details.

    """


class ExperimentalDataset(aspecd.dataset.ExperimentalDataset):
    """Entity consisting of experimental data and metadata.

    On the one hand this class extends the metadata contained in the metadata
    property. On the other hand this class extends the dataset by further
    :class:`aspecd.dataset.Data` objects containing additional variable
    parameters with independent axes useful for further analysis.

    Attributes
    ----------
    metadata : :class:`trepr.dataset.ExperimentalDatasetMetadata`
        Metadata of dataset.

    time_stamp : :class:`aspecd.dataset.Data`
        Time stamp of each individual time trace.

        Note: actual time stamp data will not be available for each file format.

    microwave_frequency : :class:`aspecd.dataset.Data`
        Microwave frequency of each individual time trace.

        Note: actual frequency data will not be available for each file format.

    """

    def __init__(self):
        super().__init__()
        # public properties
        self.metadata = ExperimentalDatasetMetadata()
        self.time_stamp = aspecd.dataset.Data()
        self.microwave_frequency = aspecd.dataset.Data()


class CalculatedDataset(aspecd.dataset.CalculatedDataset):
    """Entity consisting of calculated data and metadata.

    As the class is fully inherited from ASpecD for simple usage, see the
    ASpecD documentation of the :class:`aspecd.dataset.CalculatedDataset`
    class for details.

    """


class DatasetFactory(aspecd.dataset.DatasetFactory):
    """
    Factory for creating dataset objects based on the source provided.

    Particularly in case of recipe-driven data analysis, there is a need to
    automatically retrieve datasets using nothing more than a source string
    that can be, e.g., a path or LOI.

    The DatasetFactory operates in conjunction with a
    :class:`trepr.io.DatasetImporterFactory` to import the actual dataset.
    See the respective class documentation for more details.

    Attributes
    ----------
    importer_factory : :class:`trepr.io.DatasetImporterFactory`
        ImporterFactory instance used for importing datasets

    """

    def __init__(self):
        super().__init__()
        self.importer_factory = trepr.io.DatasetImporterFactory()

    @staticmethod
    def _create_dataset(source=''):
        return trepr.dataset.ExperimentalDataset()


class ExperimentalDatasetMetadata(aspecd.metadata.ExperimentalDatasetMetadata):
    """Metadata for an experimental TREPR dataset.

    The metadata and their structure implemented in this class are one key
    aspect of the trepr package. They provide a unified interface for all
    functionality operating on datasets. Additionally, the metadata and
    their structure are the result of more than fifteen years of practical
    experience. Defining what kind of metadata is important and needs to be
    recorded, together with metadata formats easily writable by the
    experimenters *during* recording the data requires a thorough
    understanding of both, the method and the setup(s) used.

    Each attribute is an instance of the respective subclass. For details of
    the attributes of these classes, see their respective documentation.

    Metadata can be converted to dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`, e.g., for generating
    reports using templates and template engines.

    Attributes
    ----------
    measurement : :obj:`trepr.dataset.Measurement`
        Metadata corresponding to the measurement.

    sample : :obj:`trepr.dataset.Sample`
        Metadata corresponding to the sample.

    transient : :obj:`trepr.dataset.Transient`
        Metadata corresponding to the transient.

    experiment : :obj:`trepr.dataset.Experiment`
        Metadata corresponding to the experiment.

    spectrometer : :obj:`trepr.dataset.Spectrometer`
        Metadata corresponding to the spectrometer.

    magnetic_field : :obj:`trepr.dataset.MagneticField`
        Metadata corresponding to the magnetic field.

    background : :obj:`trepr.dataset.Background`
        Metadata corresponding to the background.

    bridge : :obj:`trepr.dataset.Bridge`
        Metadata corresponding to the bridge.

    video_amplifier : :obj:`trepr.dataset.VideoAmplifier`
        Metadata corresponding to the video amplifier.

    recorder : :obj:`trepr.dataset.Recorder`
        Metadata corresponding to the recorder.

    probehead : :obj:`trepr.dataset.Probehead`
        Metadata corresponding to the probehead.

    pump : :obj:`trepr.dataset.Pump`
        Metadata corresponding to the pump.

    temperature_control : :obj:`trepr.dataset.TemperatureControl`
        Metadata corresponding to the temperature control.

    """

    def __init__(self):
        super().__init__()
        # public properties
        self.measurement = Measurement()
        self.sample = Sample()
        self.transient = Transient()
        self.experiment = Experiment()
        self.spectrometer = Spectrometer()
        self.magnetic_field = MagneticField()
        self.background = Background()
        self.bridge = Bridge()
        self.video_amplifier = VideoAmplifier()
        self.recorder = Recorder()
        self.probehead = Probehead()
        self.pump = Pump()
        self.temperature_control = TemperatureControl()


class CalculatedDatasetMetadata(aspecd.metadata.CalculatedDatasetMetadata):
    """Metadata for a calculated dataset.

    This class contains the minimal set of metadata for a dataset consisting
    of calculated data, i.e., :class:`trepr.dataset.CalculatedDataset`.

    Metadata of actual datasets should extend this class by adding
    properties that are themselves classes inheriting from
    :class:`aspecd.metadata.Metadata`.

    Metadata can be converted to dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`, e.g., for generating
    reports using templates and template engines.
    """


class Measurement(aspecd.metadata.Measurement):
    """Metadata corresponding to the measurement.

    As this class inherits from :class:`aspecd.metadata.Measurement`,
    see the documentation of the parent class for details and the full list
    of inherited attributes.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    label : :class:`str`
        Short description of the measurement.

    """

    def __init__(self, dict_=None):
        # public properties
        self.label = ''
        super().__init__(dict_=dict_)


class Sample(aspecd.metadata.Sample):
    """Metadata corresponding to the sample.

    As this class inherits from :class:`aspecd.metadata.Sample`,
    see the documentation of the parent class for details and the full list
    of inherited attributes.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    description : :class:`str`
        Description of the measured sample.

    solvent : :class:`str`
        Name of the solvent used.

    preparation : :class:`str`
        Short details of the sample preparation.

    tube : :class:`str`
        Type and dimension of the sample tube used.

    """

    def __init__(self, dict_=None):
        # public properties
        self.description = ''
        self.solvent = ''
        self.preparation = ''
        self.tube = ''
        super().__init__(dict_=dict_)


class Transient(aspecd.metadata.Metadata):
    """Metadata corresponding to the transient.

    A transient or time trace is the EPR signal intensity over time for a given
    magnetic field point. A full 2D TREPR dataset usually consists of a
    series of transients for different magnetic field points that are
    sequentially recorded.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    points : :class:`int`
        Number of measuring points.

    length : :obj:`aspecd.metadata.PhysicalQuantity`
        Object of the PhysicalQuantity class from ASpecD.

    trigger_position : :class:`int`
        Position of the trigger.

    """

    def __init__(self, dict_=None):
        # public properties
        self.points = None
        self.length = aspecd.metadata.PhysicalQuantity()
        self.trigger_position = None
        super().__init__(dict_=dict_)


class Experiment(aspecd.metadata.Metadata):
    """Metadata corresponding to the experiment.

    Two parameters are relevant for a TREPR experiment that are not
    contained in other parts of the metadata: Number of runs and the shot
    repetition rate. The latter is determined by the laser system used as
    pump source.

    Note that a TREPR experiment can be understood as a special variant of a
    pump-probe experiment with the laser excitation acting as pump and the
    EPR signal intensity recording as function of time as probe.

    Typical shot repetition rates range from 20 Hz to less than 1 Hz and are
    mostly determined not only by the laser system used but as well by the
    kinetics of the sample investigated. To prevent degradation, the sample
    should be fully returned to its ground state between two excitations.
    The TREPR signal decay can *not* be used as a valid measure of the
    kinetics and the return of the sample to its ground state, as the spin
    polarisation usually decays much faster than the sample returns to its
    ground state.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    runs : :class:`int`
        Number of recorded runs.

    shot_repetition_rate : :obj:`aspecd.metadata.PhysicalQuantity`
        Shot repetition rate of the experiment.

    """

    def __init__(self, dict_=None):
        # public properties
        self.runs = None
        self.shot_repetition_rate = aspecd.metadata.PhysicalQuantity()
        super().__init__(dict_=dict_)


class Spectrometer(aspecd.metadata.Metadata):
    """Metadata corresponding to the spectrometer.

    TREPR spectrometers almost always consist of several exchangeable
    components, even if commercially available modular systems (*e.g.*
    Bruker EMX or ELEXSYS) are used. Therefore, the spectrometer model and
    the software (including its version information) need to be recorded.

    All the other components of the (TR)EPR spectrometer, such as magnet
    system, microwave bridge and amplifier, recorder, probehead and laser
    system, as well as the temperature controller have their own respective
    classes containing the necessary information.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    model : :class:`str`
        Model of the spectrometer used.

    software : :class:`str`
        Name and version of the software used.

    """

    def __init__(self, dict_=None):
        # public properties
        self.model = ''
        self.software = ''
        super().__init__(dict_=dict_)


class MagneticField(aspecd.metadata.Metadata):
    """Metadata corresponding to the magnetic field.

    The external magnetic field is a crucial component of EPR spectroscopy,
    and the corresponding metadata contain both, information about the
    settings relevant for the experiment conducted as well as the devices (
    field probe, controller, power supply) used. The latter is crucial,
    as most EPR spectrometers (with the exception of highly integrated
    benchtop devices) are modular and consist of exchangeable components.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    field_probe_type : :class:`str`
        Type of field probe used.

        Usually, a Hall probe will be used, with a somewhat limited
        accuracy. The other option is to use a Teslameter, *i.e.* a small
        NMR device, providing much higher accuracy.

    field_probe_model : :class:`str`
        Model of field probe used.

    start : :obj:`aspecd.metadata.PhysicalQuantity`
        Start of the magnetic field sweep.

    stop: :obj:`aspecd.metadata.PhysicalQuantity`
        End of the magnetic field sweep.

    step : :obj:`aspecd.metadata.PhysicalQuantity`
        Step size of the magnetic field sweep.

    sequence : :class:`str`
        Sequence of field steps, such as "up", "down", "out", or "in".

    controller : :class:`str`
        Model of the controller used.

    power_supply : :class:`str`
        Model of the power supply used.

    """

    def __init__(self, dict_=None):
        # public properties
        self.field_probe_type = ''
        self.field_probe_model = ''
        self.start = aspecd.metadata.PhysicalQuantity()
        self.stop = aspecd.metadata.PhysicalQuantity()
        self.step = aspecd.metadata.PhysicalQuantity()
        self.sequence = ''
        self.controller = ''
        self.power_supply = ''
        super().__init__(dict_=dict_)


class Background(aspecd.metadata.Metadata):
    """Metadata corresponding to the background.

    When recording TREPR data, the laser excitation typically creates a
    field-independent background signal that needs to be subtracted from the
    data.

    There are different ways of how to measure the background signal. The
    simplest (and most often used) approach is to record sufficient baseline
    at the lower and upper magnetic field range. Other approaches are to
    record the background every Nth time trace and directly subtract this
    background signal in the transient recorder. Both methods have their
    advantages and disadvantages. For details, see the literature.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    field : :obj:`aspecd.metadata.PhysicalQuantity`
        Magnetic field position of background trace.

    occurrence : :class:`int`
        Number of time traces after which a background trace is recorded.

    polarisation: :class:`str`
        Type of background polarisation *i.e.* absorptive or emissive.

        An emissive polarisation is a clear hint that the phase is set 180° off.

    intensity : :obj:`aspecd.metadata.PhysicalQuantity`
        Amplitude of background signal.

        Sometimes, the background signal intensity can be much larger than
        the actual signal intensity, and in this case, the signal shape
        strongly depends on the stability of the setup, as a large
        background intensity gets subtracted. Therefore, knowing the
        background signal intensity is particulary important in case it is
        directly subtracted during data recording.

    """

    def __init__(self, dict_=None):
        # public properties
        self.field = aspecd.metadata.PhysicalQuantity()
        self.occurrence = None
        self.polarisation = ''
        self.intensity = aspecd.metadata.PhysicalQuantity()
        super().__init__(dict_=dict_)


class Bridge(aspecd.metadata.Metadata):
    """Metadata corresponding to the microwave bridge.

    The microwave bridge contains the microwave source and parts of the
    detection system. Therefore, the crucial experimental parameters such as
    attenuation and power, microwave frequency and detection system used are
    contained as well as the description of the devices, *i.e.* the bridge
    itself, its controller, and the frequency counter, as these can be
    different interchangeable components.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    model : :class:`str`
        Model of the microwave bridge used.

    controller : :class:`str`
        Model of the bridge controller used.

    attenuation : :obj:`aspecd.metadata.PhysicalQuantity`
        Attenuation of the microwave power in dB.

        Without knowing the unattenuated source power, the attenuation is a
        rather useless value, although it gets often used, particularly in
        lab jargon. Typical microwave bridges have source powers of 200 mW
        in X-Band, but newer devices sometimes deliver only 150 mW.

    power : :obj:`aspecd.metadata.PhysicalQuantity`
        Output power of the microwave.

        The actual output power of the microwave used for the experiment,
        *i.e.* the source power reduced by the attenuation. Typical values
        are in the range of 20 mW to 20 µW.

    detection : :class:`str`
        Type of the detection used.

        There are two types of detection: diode and mixer. The latter
        usually allows for quadrature detection, *i.e.* detecting both, the
        absorptive and dispersive signal components.

        For the best time resolution, mixer detection in conjunction with a
        video amplifier get used, whereas conventional cw-EPR bridges only
        have diode detection combined with a rather low-bandwidth preamplifier.

    frequency_counter : :class:`str`
        Model of the frequency counter used.

        Depending on the setup used, this can be included in the bridge.
        Otherwise, it will often be a HP device.

    mw_frequency : :obj:`aspecd.metadata.PhysicalQuantity`
        Microwave frequency.

        The actual microwave frequency used for the experiment. Usually,
        this is a scalar number. Depending on the experiment control
        software used, the microwave frequency for each transient will be
        recorded, thus allowing for analysing frequency drifts. This is
        particularly helpful in case of long-running experiments (12+ h).
        By comparing the amplitude of the frequency drift with the field
        step width, the potential impact in the signal shape can be directly
        calculated.

    """

    def __init__(self, dict_=None):
        # public properties
        self.model = ''
        self.controller = ''
        self.attenuation = aspecd.metadata.PhysicalQuantity()
        self.power = aspecd.metadata.PhysicalQuantity()
        self.detection = ''
        self.frequency_counter = ''
        self.mw_frequency = aspecd.metadata.PhysicalQuantity()
        super().__init__(dict_=dict_)


class VideoAmplifier(aspecd.metadata.Metadata):
    """Metadata corresponding to the video amplifier.

    Depending on the detection (diode or mixer), either a low-bandwidth
    preamplifier will be used or a high-bandwidth video amplifier. The
    latter is only true for dedicated transient or pulsed bridges.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    bandwidth : :obj:`aspecd.metadata.PhysicalQuantity`
        Bandwidth of the video amplifier.

        Depending on the specific model, values such as 20, 50, 100, and 200
        dB are typically found. Newer devices have only reduced settings.

        The bandwidth of the amplifier is one of the parameters determining
        the overall time resolution of the system, but higher bandwidth
        usually comes with higher noise levels.

    amplification : :obj:`aspecd.metadata.PhysicalQuantity`
        Amplification in dB.

    """

    def __init__(self, dict_=None):
        # public properties
        self.bandwidth = aspecd.metadata.PhysicalQuantity()
        self.amplification = aspecd.metadata.PhysicalQuantity()
        super().__init__(dict_=dict_)


class Recorder(aspecd.metadata.Metadata):
    """Metadata corresponding to the recorder.

    The recorder works as an A/D converter, recording the time-dependent EPR
    signal intensity with high time resolution.

    Typically, digital oscilloscopes are used as transient recorders for
    TREPR spectroscopy. In case of commercially available pulsed EPR setups
    (Bruker ELEXSYS), the SpecJet can be used, basically a digital
    oscilloscope as well.

    Two characteristics of a transient recorder are particularly important:
    the time resolution (determined by the bandwidth and time base), and the
    vertical resolution. While the former is rarely limiting for TREPR
    spectra, the vertical resolution is often rather an issue. While
    stand-alone digital oscilloscopes provided a vertical resolution of 11
    bit already in the 1990s, only with the advent of the Bruker SpecJet III
    the vertical resolution for commercial Bruker spectrometers has
    increased beyond 8 bit. The latter is rather useless for recording TREPR
    data, as it translates to only 256 discrete steps. Having in mind that
    typically, TREPR spectra consist of both, absorptive and emissive signal
    components, this amounts to at most only 128 discrete steps per
    direction in the signal maximum. Therefore, weaker parts of the overall
    signal are usually not well resolved in this case.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    model : :class:`str`
        Model of the recorder used.

    averages : :class:`int`
        Number of accumulations recorded.

        Typically, at least 20, but often several hundreds of accumulations
        are recorded for each individual field point to increase the
        signal-to-noise ratio.

    time_base : :obj:`aspecd.metadata.PhysicalQuantity`
        Time base of the recorder.

        The time base is the time between two data points recorded. The time
        base should be much shorter (by a factor of 5--10) than the inverse
        bandwidth of the slowest component of the signal to faithfully
        digitise the actual signal shape (Nyquist sampling theorem).

        Note: While the time base is part of the time resolution of the
        overall system, usually, the time base of the recorder is not the
        limiting characteristic, but rather either the bandwidth of the
        microwave amplifier or the probehead.

    bandwidth : :obj:`aspecd.metadata.PhysicalQuantity`
        Bandwidth of the recorder.

        Typically, bandwidth and time base are connected. For the bandwidth,
        the same applies that has been said for the time base.

    pretrigger : :obj:`aspecd.metadata.PhysicalQuantity`
        Length of time trace before the actual trigger.

        Transients should never start with the actual laser excitation,
        but rather some time before, in order to allow for compensating for
        DC offsets.

        Note that the trigger point does not necessarily allow for
        determining the true signal rise time, due to the time the
        electrical signal takes to travel through the cables. After all,
        the speed of light translates "only" to 30 cm per nanosecond,
        and with typical cable lengths of several metres, this translates
        well into tens of nanoseconds.

    coupling : :class:`str`
        Type of coupling.

        Usually, this is either DC or AC coupling.

    impedance : :obj:`aspecd.metadata.PhysicalQuantity`
        Impedance of the recorder input channel.

        The impedance setting of the recorder input should match that of the
        output of the detector. Take care of any additional filters or alike
        in the signal pathway that typically change the impedance. While
        microwave bridges usually have 50 Ohm impedance, high-pass filters
        will typically translate this to several kOhm, making it necessary
        to change the impedance of the recorder input channel to 1 MOhm.

    sensitivity : :obj:`aspecd.metadata.PhysicalQuantity`
        Sensitivity of the recorder.

        The sensitivity setting determines the vertical resolution of the
        signal. Therefore, the optimum should be sought after between
        maximum vertical resolution and not clipping the signal.

    """

    def __init__(self, dict_=None):
        # public properties
        self.model = ''
        self.averages = None
        self.time_base = aspecd.metadata.PhysicalQuantity()
        self.bandwidth = aspecd.metadata.PhysicalQuantity()
        self.pretrigger = aspecd.metadata.PhysicalQuantity()
        self.coupling = ''
        self.impedance = aspecd.metadata.PhysicalQuantity()
        self.sensitivity = aspecd.metadata.PhysicalQuantity()
        super().__init__(dict_=dict_)


class Probehead(aspecd.metadata.Metadata):
    """Metadata corresponding to the probehead.

    Often, resonating structures get used in EPR spectroscopy, but as this
    is not always the case, the term "probehead" is more generic.

    In all except of fully integrated benchtop spectrometers, the probehead
    can readily be exchanged. As each probehead has its own characteristics,
    it is crucially important to note at least type and model. The coupling
    (critically or overcoupled) determines the bandwidth of the resonator,
    and in all but pulsed experiments, usually, critical coupling is used.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    type : :class:`str`
        Type of the probehead used.

        There are several different types of probeheads regularly used. For
        resonators, there are, *e.g.*, dielectic and split-ring resonators,
        cylindrical and rectangular cavities. More special would be
        Fabry-Perot and stripline resonators. Sometimes, even resonator-free
        designs are used as probeheads.

    model : str
        Model of the probehead used.

        Commercial probeheads come with a distinct model that goes in here.
        In all other cases, use a short, memorisable, and unique name.

    coupling : :class:`str`
        Type of coupling.

        Usually either critically (default) or overcoupled

    """

    def __init__(self, dict_=None):
        # public properties
        self.type = ''
        self.model = ''
        self.coupling = ''
        super().__init__(dict_=dict_)


class Pump(aspecd.metadata.Metadata):
    """Metadata corresponding to the optical excitation..

    A TREPR experiment can be understood as a special variant of a
    pump-probe experiment with the laser excitation acting as pump and the
    EPR signal intensity recording as function of time as probe.

    This class contains all information necessary to describe the laser
    system used as a pump. This includes, besides wavelength, power,
    and repetition rate, the type of tunable (if any), and filters situated
    within the beam path.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    type : :class:`str`
        Type of the pump used.

        This typically translates to "laser", and it may be extended by the
        actual type of laser (*e.g.*, Nd:YAG).

    model : :class:`str`
        Model of the pump used.

    wavelength : :obj:`aspecd.metadata.PhysicalQuantity`
        Wavelength of the optical excitation.

        Depending on whether a tunable light source (OPO, dye laser) is
        used, this can be either rather discrete wavelengths (the harmonics
        of the fundamental wavelength of the laser used), or specific
        wavelengths (typically with an accuracy of 1 nm).

    power : :obj:`aspecd.metadata.PhysicalQuantity`
        Power of the optical excitation.

        Values are typically in the range of 0.5--5 mJ. Make sure to
        *always* record the laser power, as too high laser powers can damage
        your sample (or even your setup).

        Note that measuring the laser power is typically done outside the
        probehead. Therefore, the actual laser power incident at the sample
        is much less, depending on how well you've focussed/collimated the
        excitation beam as well as on the physical structure of the window
        in the probehead. Often, the window contains a metal grid used to
        prevent moo much micrwove leakage, but reducing the incident laser
        power on the sample further.

    repetition_rate : :obj:`aspecd.metadata.PhysicalQuantity`
        Repetition rate of the optical excitation.

        Typical values range from 20 Hz to <1 Hz. The repetition rate of the
        pump determines the shot repetition rate of the whole experiment.

        The kinetics of the sample investigated is crucial for determining
        the correct repetition rate. To prevent degradation, the sample
        should be fully returned to its ground state between two
        excitations. The TREPR signal decay can *not* be used as a valid
        measure of the kinetics and the return of the sample to its ground
        state, as the spin polarisation usually decays much faster than the
        sample returns to its ground state.

    tunable_type : :class:`str`
        Type of the tunable used.

        Usually, this is eiter "OPO" or "dye laser".

    tunable_model : :class:`str`
        Model of the tunable used.

    tunable_dye : :class:`str`
        Name of Laser dye.

        In case of a dye laser, details of the laser dye used.

    tunable_position : :class:`str`
        Position of the tunable used.

        In case of an OPO, you may provide the position of the stepper motor
        here. Please note that these values can only be used for comparision
        within one setup and one calibration.

    filter : :class:`str`
        Type of the filter(s) used.

        Often, additional filters are placed within the excitation beam.

    """

    def __init__(self, dict_=None):
        # public properties
        self.type = ''
        self.model = ''
        self.wavelength = aspecd.metadata.PhysicalQuantity()
        self.power = aspecd.metadata.PhysicalQuantity()
        self.repetition_rate = aspecd.metadata.PhysicalQuantity()
        self.tunable_type = ''
        self.tunable_model = ''
        self.tunable_dye = ''
        self.tunable_position = None
        self.filter = ''
        super().__init__(dict_=dict_)


class TemperatureControl(aspecd.metadata.TemperatureControl):
    """Metadata corresponding to the temperature control.

    As this class inherits from :class:`aspecd.metadata.TemperatureControl`,
    see the documentation of the parent class for details and the full list
    of inherited attributes.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    Attributes
    ----------
    cryostat : :class:`str`
        Model of the cryostat used.

    cryogen : :class:`str`
        Cryogen used.

        Typically, this is either LN2 (for temperatures down to 80K) or LHe
        (for temperatures down to 4 K)

    """

    def __init__(self, dict_=None):
        # public properties
        self.cryostat = ''
        self.cryogen = ''
        super().__init__(dict_=dict_)
